/////////////////////////////////////////////////////////////////////////////
//
//	$Header: /WinArena/Raster.h 5     12/30/05 1:11a Lee $
//
//	File: Raster.h
//
//
/////////////////////////////////////////////////////////////////////////////


#pragma once


#include "ParseBSA.h"
#include "Animation.h"


#define FloatToFixed(x)		(long((x) * float(c_FixedPointOne)))
#define IntToFixed(x)		((x) << c_FixedPointBits)
#define FixedFloor(x)		((x) & c_FixedPointWhole)
#define FixedCeil(x)		(((x) + c_FixedPointMask) & c_FixedPointWhole)


#define c_FixedPointBits	15
#define c_FixedPointMask	0x00007FFF
#define c_FixedPointWhole	0xFFFF8000
#define c_FixedPointOne		0x00008000
#define c_FixedPointHalf	0x00004000

// Extra bits used to store inverse Z value to increase precision.
#define c_ZDepthBias		4

#define c_ZNear				(c_FixedPointOne >> 2)
#define c_ZFar				(32 << c_FixedPointBits)
#define c_ZRange			(zFar - zNear)

#define c_WallHeight		(IntToFixed(3) >> 2)
#define c_WallLength		(IntToFixed(5) >> 2)
#define c_EyeLevel			((c_WallHeight * 35) >> 6)
#define c_BulletLevel		((c_WallHeight * 30) >> 6)

#define c_ZScale			(IntToFixed(6) >> 2)


/////////////////////////////////////////////////////////////////////////////
//
struct Vertex_t
{
	long x;
	long y;
	long z;
	long u;
	long v;
	long l;
};


class CwaRaster
{
private:
	BYTE*	m_pFrame;				// frame buffer, storing 32-bit ARGB colors
	DWORD*	m_pPalette;				// color palette to use when displaying current frame
	BYTE*	m_pLightMap;			// current lightmap, used to fade pixel colors towards black or fog color
	DWORD*	m_pShapeID;				// ID buffer, stores ID value that maps back to the object rendered there
	long*	m_pDepth;				// depth buffer, storing 1/z
	DWORD	m_ViewWidth;			// size of rendering area; this exclude the HUD at the bottom of the screen
	DWORD	m_ViewHeight;
	DWORD	m_FrameHeight;			// size of frame buffer
	DWORD	m_FrameWidth;
	DWORD	m_WindowWidth;			// resolution of the window where frame will be blitted
	DWORD	m_WindowHeight;
	DWORD	m_PixelCount;
	DWORD	m_BlitMode;

	CwaTexture	m_DefaultTexture;
	CwaTexture*	m_pTexture;

public:
	CwaRaster(void);
	~CwaRaster(void);

	void Allocate(DWORD width, DWORD height);
	void SetWindowSize(DWORD windowWidth, DWORD windowHeight);
	void SetBlitMode(DWORD mode);
	void ClearFrameBuffer(DWORD color);
	void ScreenShot(char filename[]);

	void DisplayFrame(HDC hDC);

	void DrawBox(long x, long y, long width, long height, BYTE color);

	void Line(long x0, long y0, long x1, long y1, BYTE color);
	void Circle(long x0, long y0, long radius, BYTE color);
	void FilledCircle(long x0, long y0, long radius, BYTE color);
	void Polygon(long pX[], long pY[], DWORD count, BYTE color);

	void Blit(const CwaTexture &texture, DWORD posX, DWORD posY);
	void BlitRect(const CwaTexture &texture, DWORD posX, DWORD posY, DWORD dx, DWORD dy, DWORD width, DWORD height);
	void BlitTransparent(const CwaTexture &texture, DWORD posX, DWORD posY);
	void BlitTransparentMirror(const CwaTexture &texture, DWORD posX, DWORD posY);

	void DrawTriangleSolid(Vertex_t *pA, Vertex_t *pB, Vertex_t *pC, BYTE color);
	void DrawTriangle(Vertex_t *pA, Vertex_t *pB, Vertex_t *pC, DWORD shapeID);
	void DrawTriangleTransparent(Vertex_t *pA, Vertex_t *pB, Vertex_t *pC, DWORD shapeID);
	void DrawTrianglePerspective(Vertex_t *pA, Vertex_t *pB, Vertex_t *pC, DWORD shapeID);

	void DrawSprite(Vertex_t corners[], DWORD shapeID);
	void DrawSpriteMirror(Vertex_t corners[], DWORD shapeID);
	void DrawSpriteTranslucent(Vertex_t corners[], DWORD shapeID);
	void DrawSpriteTranslucentMirror(Vertex_t corners[], DWORD shapeID);

	void  DrawNumber(FontElement_t font[], DWORD x, DWORD y, long value, DWORD color);
	void  DrawSignedNumber(FontElement_t font[], DWORD x, DWORD y, long value, DWORD color);
	void  DrawString(FontElement_t font[], DWORD x, DWORD y, char message[], DWORD color);
	DWORD StringLength(FontElement_t font[], char message[]);

	void  SetLightMap(BYTE *pLightMap);
	void  SetPalette(DWORD *pPalette);
	void  SetTexture(CwaTexture *pTexture);

	DWORD GetShapeID(DWORD x, DWORD y)		{ return m_pShapeID[(y * m_FrameWidth) + x]; }
	long  GetPixelDepth(DWORD x, DWORD y)	{ return m_pDepth[(y * m_FrameWidth) + x]; }

	DWORD CountNonzeroPixels(void);
	DWORD PixelCount(void)					{ return m_PixelCount; }

	DWORD FrameWidth(void)		{ return m_FrameWidth; }
	DWORD FrameHeight(void)		{ return m_FrameHeight; }
};


long FixedSqrt(DWORD a);
long FixedDistance(long x, long z);


_forceinline
//inline
long FixedMul(long a, long b)
{
	__asm {
		mov  eax,a
		imul b
		shrd eax,edx,c_FixedPointBits
	}

	// eax is implicitly returned
}

_forceinline
//inline
long FixedDiv(long a, long b)
{
//	if (0 != b) {
		if ((a^b)&0x8000000) {
			__asm {
				mov  eax,a
				mov  ebx,b
				neg  eax
				xor  edx,edx
				shld edx,eax,c_FixedPointBits
				sal  eax,c_FixedPointBits
				idiv ebx
				neg  eax
			}
		}
		else {
			__asm {
				mov  eax,a
				mov  ebx,b
				xor  edx,edx
				shld edx,eax,c_FixedPointBits
				sal  eax,c_FixedPointBits
				idiv ebx
			}
		}
//	}
//	else {
//		__asm {
//			mov eax,0x7FFFFFFF
//		}
//	}

	// eax is implicitly returned
}

//#define FixedMul(r,a,b)   __asm mov eax,a   __asm mul b   __asm shrd eax,edx,16  __asm mul r,eax

