/////////////////////////////////////////////////////////////////////////////
//
//	$Header: /WinArena/WinArena.cpp 8     12/31/05 12:45a Lee $
//
//	File: WinArena.cpp
//
//
//	Defines the entry point for the application.
//
/////////////////////////////////////////////////////////////////////////////


#include "WinArena.h"
#include "Resource.h"
#include "ParseBSA.h"
#include "Raster.h"
#include "SoundDriver.h"
#include "World.h"


// Global Variables:
HINSTANCE	g_hInstance			= NULL;		// current instance
HWND		g_hWindow			= NULL;
HMENU		g_hMenu				= NULL;
TCHAR		g_Title[128];					// The title bar text
TCHAR		g_WindowClass[128];				// the main window class name
DWORD		g_WindowWidth		= 640;
DWORD		g_WindowHeight		= 400;
CwaRaster*	g_pRaster			= NULL;
CwaWorld*	g_pWorld			= NULL;
bool		g_Minimized			= false;
bool		g_TrackingMouse		= false;	// need to track when mouse is in client window
bool		g_LeftMouse			= false;
bool		g_RightMouse		= false;
bool		g_ShowFPS			= true;
bool		g_Illumination		= true;

bool		g_KeyDownUp			= false;
bool		g_KeyDownLeft		= false;
bool		g_KeyDownDown		= false;
bool		g_KeyDownRight		= false;


/////////////////////////////////////////////////////////////////////////////
//
//	ResetWindowSize()
//
void ResetWindowSize(DWORD newWidth, DWORD newHeight)
{
	g_WindowWidth  = newWidth;
	g_WindowHeight = newHeight;

	RECT clientRect, windowRect, desktopRect;

	GetClientRect(g_hWindow, &clientRect);
	GetWindowRect(g_hWindow, &windowRect);
	GetWindowRect(GetDesktopWindow(), &desktopRect);

	long width  = g_WindowWidth  + (windowRect.right  - windowRect.left) - (clientRect.right  - clientRect.left);
	long height = g_WindowHeight + (windowRect.bottom - windowRect.top)  - (clientRect.bottom - clientRect.top);

	long x = (desktopRect.right  - desktopRect.left) - width;
	long y = (desktopRect.bottom - desktopRect.top)  - height;

	SetWindowPos(g_hWindow, NULL, x/2, y/2, width, height, SWP_NOZORDER);
}


/////////////////////////////////////////////////////////////////////////////
//
//	WndProc()
//
//	Processes messages for the main window.
//
LRESULT CALLBACK WndProc(HWND hWindow, UINT message, WPARAM wParam, LPARAM lParam)
{
	int wmId, wmEvent;
	PAINTSTRUCT ps;
	HDC hDC;

	switch (message) {
		case WM_COMMAND:
			wmId    = LOWORD(wParam);
			wmEvent = HIWORD(wParam);

			// Parse the menu selections:
			switch (wmId) {
				case ID_FILE_SCREENSHOT:
					if (NULL != g_pRaster) {
						g_pRaster->ScreenShot("screen.gif");
					}
					break;

				case ID_QUALITY_SHOWFPS:
					g_ShowFPS = !g_ShowFPS;
					if (NULL != g_pWorld) {
						g_pWorld->ShowFPS(g_ShowFPS);
					}
					CheckMenuItem(g_hMenu, ID_QUALITY_SHOWFPS, g_ShowFPS ? MF_CHECKED : MF_UNCHECKED);
					break;

				case ID_QUALITY_ILLUMINATE:
					g_Illumination = !g_Illumination;
					if (NULL != g_pWorld) {
						g_pWorld->EnablePlayerIllumination(g_Illumination);
					}
					CheckMenuItem(g_hMenu, ID_QUALITY_ILLUMINATE, g_Illumination ? MF_CHECKED : MF_UNCHECKED);
					break;

				case ID_QUALITY_RESETWINDOWTO320X200:
					ResetWindowSize(320, 200);
					break;

				case ID_QUALITY_RESETWINDOWTO640X400:
					ResetWindowSize(640, 400);
					break;

				case ID_QUALITY_RESETWINDOWTO960X600:
					ResetWindowSize(960, 600);
					break;

				case ID_QUALITY_320X200:
					if (NULL != g_pRaster) {
						g_pRaster->Allocate(320, 200);
						g_pRaster->SetWindowSize(g_WindowWidth, g_WindowHeight);
					}
					if (NULL != g_pWorld) {
						g_pWorld->SetResolution(320, 200);
					}
					CheckMenuItem(g_hMenu, ID_QUALITY_320X200, MF_CHECKED);
					CheckMenuItem(g_hMenu, ID_QUALITY_640X400, MF_UNCHECKED);
					CheckMenuItem(g_hMenu, ID_QUALITY_960X600, MF_UNCHECKED);
					break;

				case ID_QUALITY_640X400:
					if (NULL != g_pRaster) {
						g_pRaster->Allocate(640, 400);
						g_pRaster->SetWindowSize(g_WindowWidth, g_WindowHeight);
					}
					if (NULL != g_pWorld) {
						g_pWorld->SetResolution(640, 400);
					}
					if (g_WindowWidth < 640) {
						ResetWindowSize(640, 400);
					}
					CheckMenuItem(g_hMenu, ID_QUALITY_320X200, MF_UNCHECKED);
					CheckMenuItem(g_hMenu, ID_QUALITY_640X400, MF_CHECKED);
					CheckMenuItem(g_hMenu, ID_QUALITY_960X600, MF_UNCHECKED);
					break;

				case ID_QUALITY_960X600:
					if (NULL != g_pRaster) {
						g_pRaster->Allocate(960, 600);
						g_pRaster->SetWindowSize(g_WindowWidth, g_WindowHeight);
					}
					if (NULL != g_pWorld) {
						g_pWorld->SetResolution(960, 600);
					}
					if (g_WindowWidth < 960) {
						ResetWindowSize(960, 600);
					}
					CheckMenuItem(g_hMenu, ID_QUALITY_320X200, MF_UNCHECKED);
					CheckMenuItem(g_hMenu, ID_QUALITY_640X400, MF_UNCHECKED);
					CheckMenuItem(g_hMenu, ID_QUALITY_960X600, MF_CHECKED);
					break;

				case ID_QUALITY_FASTBLIT:
					if (NULL != g_pRaster) {
						g_pRaster->SetBlitMode(BLACKONWHITE);
					}
					CheckMenuItem(g_hMenu, ID_QUALITY_FASTBLIT,   MF_CHECKED);
					CheckMenuItem(g_hMenu, ID_QUALITY_SMOOTHBLIT, MF_UNCHECKED);
					break;

				case ID_QUALITY_SMOOTHBLIT:
					if (NULL != g_pRaster) {
						g_pRaster->SetBlitMode(HALFTONE);
					}
					CheckMenuItem(g_hMenu, ID_QUALITY_FASTBLIT,   MF_UNCHECKED);
					CheckMenuItem(g_hMenu, ID_QUALITY_SMOOTHBLIT, MF_CHECKED);
					break;

				case ID_DUMP_BSAINDEX:
					DumpBSAIndex();
					break;

				case ID_DUMP_ANIMATIONS:
					DumpDFA();
					DumpCIF();
					break;

				case ID_DUMP_CREATURES:
					DumpCFA();
					break;

				case ID_DUMP_IMAGES:
					DumpImages();
					break;

				case ID_DUMP_INFO:
					DumpInfo();
					break;

				case ID_DUMP_MAPS:
					DumpMaps();
					break;

				case ID_DUMP_SOUNDS:
					DumpSounds();
					break;

				case ID_DUMP_TEXTURES:
					DumpWalls();
					break;

				case IDM_EXIT:
					DestroyWindow(hWindow);
					break;

				default:
					return DefWindowProc(hWindow, message, wParam, lParam);
			}
			break;

		case WM_DESTROY:
			PostQuitMessage(0);
			break;

		case WM_ERASEBKGND:
			return TRUE;
			break;

		case WM_PAINT:
			hDC = BeginPaint(hWindow, &ps);
			if (NULL != g_pRaster) {
				g_pRaster->DisplayFrame(hDC);
			}
			// TODO: Add any drawing code here...
			EndPaint(hWindow, &ps);
			break;

		case WM_KEYDOWN:
			if (VK_LEFT == wParam) {
				g_KeyDownLeft = true;
			}
			else if (VK_UP == wParam) {
				g_KeyDownUp = true;
			}
			else if (VK_DOWN == wParam) {
				g_KeyDownDown = true;
			}
			else if (VK_RIGHT == wParam) {
				g_KeyDownRight = true;
			}
			else if (VK_ESCAPE == wParam) {
				g_pWorld->ToggleControlPanel();
			}
			break;

		case WM_KEYUP:
			if (VK_LEFT == wParam) {
				g_KeyDownLeft = false;
			}
			else if (VK_UP == wParam) {
				g_KeyDownUp = false;
			}
			else if (VK_DOWN == wParam) {
				g_KeyDownDown = false;
			}
			else if (VK_RIGHT == wParam) {
				g_KeyDownRight = false;
			}
			break;

		case WM_LBUTTONDOWN:
			if (NULL != g_pWorld) {
				g_pWorld->LeftMouseClick();
			}
			if (false == g_RightMouse) {
				SetCapture(g_hWindow);
			}
			g_LeftMouse = true;
			break;

		case WM_LBUTTONUP:
			if (NULL != g_pWorld) {
				g_pWorld->RightMouseClick(false);
			}
			if (false == g_RightMouse) {
				ReleaseCapture();
			}
			g_LeftMouse = false;
			break;

		case WM_RBUTTONDOWN:
		case WM_RBUTTONDBLCLK:
			if (NULL != g_pWorld) {
				g_pWorld->RightMouseClick(true);
			}
			if (false == g_LeftMouse) {
				SetCapture(g_hWindow);
			}
			g_RightMouse = true;
			break;

		case WM_RBUTTONUP:
			if (NULL != g_pWorld) {
				g_pWorld->RightMouseClick(false);
			}
			if (false == g_LeftMouse) {
				ReleaseCapture();
			}
			g_RightMouse = false;
			break;

		case WM_MOUSEMOVE:
			if ((NULL != g_pWorld) && (NULL != g_pRaster)) {
				DWORD x = LOWORD(lParam);
				DWORD y = HIWORD(lParam);

				if (short(x) < 0) {
					x = 0;
				}
				else if (x >= g_WindowWidth) {
					x = g_WindowWidth - 1;
				}

				if (short(y) < 0) {
					y = 0;
				}
				else if (y >= g_WindowHeight) {
					y = g_WindowHeight - 1;
				}

				// Transform the mouse position from window coordinates to
				// frame buffer coordinates.
				x = (x * g_pRaster->FrameWidth())  / g_WindowWidth;
				y = (y * g_pRaster->FrameHeight()) / g_WindowHeight;

				g_pWorld->SetMousePosition(x, y);
			}

			// If the mouse is not being tracked, turn tracking back on, and
			// hide the mouse pointer (cursor).  The rasterizer will handle
			// drawing a game-mode dependent pointer through the frame buffer,
			// so we don't want the Windows pointer to also be visible at the
			// same time.
			//
			if (false == g_TrackingMouse) {
				g_TrackingMouse = true;
				TRACKMOUSEEVENT tracker;
				tracker.cbSize      = sizeof(tracker);
				tracker.dwFlags     = TME_LEAVE;
				tracker.hwndTrack   = g_hWindow;
				tracker.dwHoverTime = HOVER_DEFAULT;
				TrackMouseEvent(&tracker);
				ShowCursor(FALSE);
			}
			return DefWindowProc(hWindow, message, wParam, lParam);

		case WM_MOUSELEAVE:
			//
			// Once we receive WM_MOUSELEAVE, tracking is turned off.
			// Have to explicitly call TrackMouseEvent() again to turn
			// it back one.  This will be done in WM_MOUSEMOVE.
			//
			// Need to do this to turn the mouse cursor on again when it
			// moves out of the client area.  This allows the mouse cursor
			// to be visible if the user is trying to click on a menu
			// button or the close button.
			//
			if (g_TrackingMouse) {
				g_TrackingMouse = false;
				ShowCursor(TRUE);
			}
			break;

		case WM_SIZE:
			if (SIZE_MINIMIZED != wParam) {
				g_Minimized    = false;
				g_WindowWidth  = LOWORD(lParam);
				g_WindowHeight = HIWORD(lParam);
				if (NULL != g_pRaster) {
					g_pRaster->SetWindowSize(g_WindowWidth, g_WindowHeight);
				}
			}
			else {
				g_Minimized = true;
			}
			break;

		default:
			return DefWindowProc(hWindow, message, wParam, lParam);
	}

	return 0;
}


/////////////////////////////////////////////////////////////////////////////
//
//	MyRegisterClass()
//
//	This function and its usage are only necessary if you want this code to
//	be compatible with Win32 systems prior to the 'RegisterClassEx' function
//	that was added to Windows 95. It is important to call this function so
//	that the application will get 'well formed' small icons associated with it.
//
ATOM MyRegisterClass(HINSTANCE hInstance)
{
	WNDCLASSEX wcex;

	wcex.cbSize = sizeof(WNDCLASSEX); 

	wcex.style			= CS_HREDRAW | CS_VREDRAW;
	wcex.lpfnWndProc	= (WNDPROC)WndProc;
	wcex.cbClsExtra		= 0;
	wcex.cbWndExtra		= 0;
	wcex.hInstance		= hInstance;
	wcex.hIcon			= LoadIcon(hInstance, (LPCTSTR)IDI_WINARENA);
	wcex.hCursor		= LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground	= (HBRUSH)(COLOR_WINDOW+1);
	wcex.lpszMenuName	= (LPCTSTR)IDC_WINARENA;
	wcex.lpszClassName	= g_WindowClass;
	wcex.hIconSm		= LoadIcon(wcex.hInstance, (LPCTSTR)IDI_SMALL);

	return RegisterClassEx(&wcex);
}


/////////////////////////////////////////////////////////////////////////////
//
//	InitInstance()
//
//	Saves instance handle and creates main window
//
//	In this function, we save the instance handle in a global variable and
//	create and display the main program window.
//
BOOL InitInstance(HINSTANCE hInstance, int nCmdShow)
{
	// Store instance handle in our global variable
	g_hInstance = hInstance;

	g_hWindow = CreateWindow(g_WindowClass, g_Title, WS_OVERLAPPEDWINDOW,
		CW_USEDEFAULT, 0, CW_USEDEFAULT, 0, NULL, NULL, hInstance, NULL);

	if (NULL == g_hWindow) {
		return FALSE;
	}


	ResetWindowSize(g_WindowWidth, g_WindowHeight);

	ShowWindow(g_hWindow, nCmdShow);
	UpdateWindow(g_hWindow);

	return TRUE;
}


/////////////////////////////////////////////////////////////////////////////
//
//	WinMain()
//
int APIENTRY WinMain(HINSTANCE hInstance,
                     HINSTANCE /*hPrevInstance*/,
                     LPTSTR    pCommandLine,
                     int       showCommand)
{

	// Initialize global strings
	LoadString(hInstance, IDS_APP_TITLE, g_Title, ArraySize(g_Title));
	LoadString(hInstance, IDC_WINARENA, g_WindowClass, ArraySize(g_WindowClass));
	MyRegisterClass(hInstance);

	// Perform application initialization:
	if (!InitInstance(hInstance, showCommand)) {
		return FALSE;
	}

	HACCEL hAccelTable = LoadAccelerators(hInstance, (LPCTSTR)IDC_WINARENA);

	if ((NULL != pCommandLine) && ('\0' != pCommandLine[0])) {
		strcpy(g_ArenaPath, pCommandLine);
	}
	else {
		strcpy(g_ArenaPath, ".\\");
	}

	g_hMenu = GetMenu(g_hWindow);

	CheckMenuItem(g_hMenu, ID_QUALITY_SHOWFPS, g_ShowFPS ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(g_hMenu, ID_QUALITY_ILLUMINATE, g_Illumination ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(g_hMenu, ID_QUALITY_640X400, MF_CHECKED);
	CheckMenuItem(g_hMenu, ID_QUALITY_FASTBLIT, MF_CHECKED);

	CwaRaster raster;
	raster.Allocate(640, 400);
	raster.SetWindowSize(g_WindowWidth, g_WindowHeight);

	CwaWorld world;
	world.SetRaster(&raster);
	world.SetResolution(640, 400);
	world.ShowFPS(g_ShowFPS);
	world.EnablePlayerIllumination(g_Illumination);
	world.Load();

	g_pRaster = &raster;
	g_pWorld  = &world;


	// Main message loop:
	MSG  message;
	BOOL newMessage = true;

	timeBeginPeriod(1);
	DWORD prevTime   = timeGetTime();
	DWORD fpsMarker  = prevTime;
	DWORD frameCount = 0;

	do {
		do {
			newMessage = PeekMessage(&message, NULL, 0, 0, PM_NOREMOVE);

			if (newMessage && (WM_QUIT != message.message)) {
				if (GetMessage(&message, NULL, 0, 0)) {
					if (!TranslateAccelerator(message.hwnd, hAccelTable, &message)) {
						TranslateMessage(&message);
						DispatchMessage(&message);
					}
				}
			}

		} while (newMessage && (WM_QUIT != message.message));

		if ((false == g_Minimized) && (WM_QUIT != message.message)) {
			DWORD curTime   = timeGetTime();
			DWORD timeDelta = curTime - prevTime;

			// Protect against really slow frame rates.  These arise whenever
			// accessing a menu option.  Large differences can result in huge
			// changes to the character position, which breaks the simple
			// collision detection/response code.
			if (timeDelta > 250) {
				timeDelta = 250;
			}

			if (g_KeyDownLeft) {
				world.TurnPlayer(timeDelta, true);
			}
			else if (g_KeyDownRight) {
				world.TurnPlayer(timeDelta, false);
			}

			if (g_KeyDownUp) {
				world.MovePlayer(timeDelta, true);
			}
			else if (g_KeyDownDown) {
				world.MovePlayer(timeDelta, false);
			}

			// If none of the movement buttons are pressed, but the mouse
			// button is pressed, check to see if this should move the player.
			if (!g_KeyDownLeft && !g_KeyDownRight && !g_KeyDownUp && !g_KeyDownDown && g_LeftMouse) {
				world.CheckMouseMovement(timeDelta);
			}

			world.AdvanceTime(timeDelta);

			world.Render();
			HDC hDC = GetDC(g_hWindow);
			g_pRaster->DisplayFrame(hDC);
			ReleaseDC(g_hWindow, hDC);

			prevTime = curTime;

			++frameCount;

			if ((curTime - fpsMarker) > 1000) {
				world.SetFPS(frameCount);
				fpsMarker  = curTime;
				frameCount = 0;
			}

			Sleep(0);
		}
	} while (WM_QUIT != message.message);

	g_pRaster = NULL;
	g_pWorld  = NULL;

	world.Unload();

	return int(message.wParam);
}


