/////////////////////////////////////////////////////////////////////////////
//
//	$Header: /WinArena/World.h 6     12/31/05 12:34a Lee $
//
//	File: World.h
//
//
/////////////////////////////////////////////////////////////////////////////


#pragma once


#include "Animation.h"
#include "Raster.h"
#include "SoundDriver.h"
#include "Types.h"


#define c_PortalOpenTime	1400		// milliseconds required for door/portcullis to open


class CwaWorld
{
private:
	CwaSoundDriver	m_SoundDriver;

	CwaRaster*		m_pRaster;

	bool			m_UseFogMap;
	BYTE			m_LightMapNormal[18*256];
	BYTE			m_LightMapFog[18*256];

	CwaTexture*		m_pCurrentMousePointer;
	bool			m_MouseMoveEnabled;
	long			m_MouseTurnRate;
	long			m_MouseWalkRate;

	SoundClip_t		m_SoundClips[Sound_ArraySize];
	CwaTexture		m_TextureList[Texture_ArraySize];
	CwaAnimation	m_Animations[Animation_ArraySize];
	CwaCreature		m_Creatures[128];
	CwaAnimation	m_Water;
	CwaAnimation	m_WeaponAxe;
	CwaAnimation	m_WeaponHammer;
	CwaAnimation	m_WeaponHand;
	CwaAnimation	m_WeaponMace;
	CwaAnimation	m_WeaponStaff;
	CwaAnimation	m_WeaponStar;
	CwaAnimation	m_WeaponSword;
	CwaAnimation	m_MousePointer;

	CwaAnimation	m_SpellAcid;
	CwaAnimation	m_SpellFire;
	CwaAnimation	m_SpellIce;
	CwaAnimation	m_SpellMagic;
	CwaAnimation	m_SpellShock;
	CwaAnimation	m_SpellSmoke;

	CastingSpell_t	m_SpellAnimation;
	DWORD			m_SpellCastBaseTime;

	AttackState_t	m_AttackState;
	DWORD			m_AttackStartTime;
	bool			m_AttackAnimating;

	CellNode_t		m_Nodes[40][40];

	Critter_t		m_Critter[128];
	Sprite_t		m_Sprite[128];
	Projectile_t	m_Projectile[16];
	LightSource_t	m_Lights[32];
	Trigger_t		m_Triggers[64];

	DWORD			m_TriggerCount;

	Character_t		m_Character;

	long			m_Position[3];
	long			m_Direction;
	long			m_DirectionSine;
	long			m_DirectionCosine;
	DWORD			m_MaxEast;
	DWORD			m_MaxSouth;
	DWORD			m_WalkDistance;			// distance walked since last footstep sound
	bool			m_LeftFootStep;			// was most recent footstep sound left or right?
	long			m_HalfWidth;
	long			m_HalfHeight;
	long			m_FieldOfViewX;
	long			m_FieldOfViewY;
	long			m_InverseFieldOfViewX;
	long			m_InverseFieldOfViewY;
	bool			m_ShowFPS;
	bool			m_IlluminatePlayer;
	DWORD			m_FPS;
	DWORD			m_MouseX;
	DWORD			m_MouseY;
	bool			m_RightMouseDown;
	DWORD			m_RightMouseX;
	DWORD			m_RightMouseY;
	DWORD			m_BaseAttackTime;
	DWORD			m_RandomSoundDelay;
	DWORD			m_CurrentTimeDelta;
	DWORD			m_CurrentTime;
	DisplayMode_t	m_DisplayMode;
	bool			m_FullBright;			// render everything at full brightness, enabled by "Light" spell
	PauseState_t	m_PauseState;			// game is paused, usually waiting for user input

	bool			m_SelectionCompleted;	// flag gets set to true when user has picked item from list
	DWORD			m_SelectionCount;		// number of items in selection list
	DWORD			m_SelectionIndex;		// index of currently highlighted item
	DWORD			m_SelectionOffset;		// index of first visible item
	char*			m_SelectionList[32];

	Triangle_t		m_TriangleList[64];
	Triangle_t*		m_pInputTriangleQueue;
	Triangle_t*		m_pOutputTriangleQueue;
	Triangle_t*		m_pFreeTriangleQueue;

	char			m_MessageBuffer[256];
	DWORD			m_MessageBaseTime;
	DWORD			m_MessageWidth;

	inline LerpVertex(const Vertex_t &v1, const Vertex_t &v2, Vertex_t &result, long mul)
	{
		result.x = FixedMul((v2.x - v1.x), mul) + v1.x;
		result.y = FixedMul((v2.y - v1.y), mul) + v1.y;
		result.z = FixedMul((v2.z - v1.z), mul) + v1.z;
		result.u = FixedMul((v2.u - v1.u), mul) + v1.u;
		result.v = FixedMul((v2.v - v1.v), mul) + v1.v;
		result.l = FixedMul((v2.l - v1.l), mul) + v1.l;
	}

public:
	CwaWorld(void);
	~CwaWorld(void);

	bool  Load(void);
	void  Unload(void);

	void  ResampleAudio(BYTE input[], short output[], DWORD countIn, DWORD countOut, DWORD filterSize);

	void  ShowFPS(bool enable);
	void  EnablePlayerIllumination(bool enable);
	void  SetFPS(DWORD fps);
	void  SetResolution(DWORD width, DWORD height);
	void  SetRaster(CwaRaster *pRaster);

	void  SetMousePosition(DWORD x, DWORD y);
	void  LeftMouseClick(void);
	void  RightMouseClick(bool down);

	void  AdvanceTime(DWORD delta);
	void  ToggleControlPanel(void);

	bool  CheckCritterCollision(long x, long z, DWORD &id);

	bool  CellIsRenderable(DWORD east, DWORD south);
	bool  CellIsSolid(DWORD east, DWORD south);
	bool  CellIsVisible(DWORD east, DWORD south);
	bool  CellHasCeiling(DWORD east, DWORD south);
	bool  CellHasFloor(DWORD east, DWORD south);
	DWORD CellTrigger(DWORD east, DWORD south);

	void  TurnPlayer(DWORD milliseconds, bool left);
	void  TurnPlayer(long angle);
	void  SetPlayerDirection(long direction);
	void  MovePlayer(DWORD milliseconds, bool forwards);
	void  MovePlayer(long distance);
	void  CheckMouseMovement(DWORD milliseconds);
	void  PushPlayerBack(long east, long south, long minDistance);

	void  ClickOnID(DWORD pointID);
	void  SetDisplayMessage(void);

	void  ResetTriangleQueues(void);
	void  SwapInputOutputTriangleQueues(void);

	Triangle_t* NewTriangle(void);
	void        FreeTriangle(Triangle_t *pTriangle);
	Triangle_t* PopInputTriangle(void);
	void        PushOutputTriangle(Triangle_t *pTriangle);
	Triangle_t* PopOutputTriangle(void);

	long  ComputeBrightness(long position[]);

	void  UpdateLightSources(void);
	void  UpdateMouseMoving(DWORD scale);
	void  UpdateTriggers(void);

	void  ShowArmorRatings(DWORD scale);

	void  Render(void);
	void  RenderAutomap(DWORD scale);
	void  RenderCharacter(DWORD scale);
	void  RenderControlPanel(DWORD scale);
	void  RenderGameScreen(DWORD scale);
	void  RenderInventory(DWORD scale);
	void  RenderJournal(DWORD scale);
	void  RenderLoadSave(DWORD scale);
	void  RenderSpellBook(DWORD scale);
	void  RenderSpellPage(DWORD scale);
	void  RenderSpellCasting(DWORD scale);
	void  RenderWater(DWORD scale);
	void  RenderWeaponOverlay(DWORD scale);

	void  RenderCell(DWORD south, DWORD east);
	void  RenderQuad(Vertex_t corners[], DWORD shapeID, bool transparent = false);
	void  RenderSprite(DWORD spriteNum);
	void  RenderProjectile(DWORD projectileNum);
	void  RenderCritter(DWORD critterNum);
	void  RenderTriangleList(bool transparent, DWORD shapeID);

	void  ClipNearZ(void);

	void  TransformPoint(long point[]);
	bool  TransformVertices(Vertex_t corners[], DWORD count);


	/////////////////////////////////////////////////////////////////////////
	//
	//	The following methods are implemented in WorldBuilder.cpp
	//

	void  ResetLevel(void);
	void  SetCellTrigger(DWORD east, DWORD south, DWORD type);
	long  PortalAngle(Trigger_t &trigger);
	void  BuildTestLevel(void);
	void  BuildTestCity(void);
	void  BuildDemoDungeon(void);
};


