/////////////////////////////////////////////////////////////////////////////
//
//	$Header: /WinArena/WorldBuilder.cpp 2     12/30/05 1:16a Lee $
//
//	File: WorldBuilder.cpp
//
//
/////////////////////////////////////////////////////////////////////////////


#include "WinArena.h"
#include "World.h"


extern AnimInfo_t    g_AnimInfo[];
extern CritterInfo_t g_CritterInfo[];


/////////////////////////////////////////////////////////////////////////////
//
//	ResetLevel()
//
void CwaWorld::ResetLevel(void)
{
	memset(m_Sprite, 0, sizeof(m_Sprite));
	memset(m_Lights, 0, sizeof(m_Lights));
	memset(m_Projectile, 0, sizeof(m_Projectile));

	m_TriggerCount = 0;
}


/////////////////////////////////////////////////////////////////////////////
//
//	SetCellTrigger()
//
void CwaWorld::SetCellTrigger(DWORD east, DWORD south, DWORD type)
{
	if (m_TriggerCount < ArraySize(m_Triggers)) {
		DWORD triggerNum = m_TriggerCount++;

		m_Triggers[triggerNum].Flags	= type;
		m_Triggers[triggerNum].East		= east;
		m_Triggers[triggerNum].South	= south;
		m_Triggers[triggerNum].BaseTime	= 0;

		m_Nodes[south][east].Flags   &= ~CellFlag_Solid;
		m_Nodes[south][east].Flags   |= type;
		m_Nodes[south][east].Trigger  = triggerNum;
	}
}


/////////////////////////////////////////////////////////////////////////////
//
//	PortalAngle()
//
long CwaWorld::PortalAngle(Trigger_t &trigger)
{
	long angle = 0;

	if (TriggerFlag_Open & trigger.Flags) {
		angle = c_FixedPointOne;
	}
	else if (TriggerFlag_Opening & trigger.Flags) {
		angle = ((m_CurrentTime - trigger.BaseTime) * c_FixedPointOne) / c_PortalOpenTime;
	}
	else if (TriggerFlag_Closing & trigger.Flags) {
		long delta = c_PortalOpenTime - (m_CurrentTime - trigger.BaseTime);

		angle = (delta * c_FixedPointOne) / c_PortalOpenTime;
	}

	if (angle < 0) {
		angle = 0;
	}
	else if (angle > c_FixedPointOne) {
		angle = c_FixedPointOne;
	}

	return angle;
}


/////////////////////////////////////////////////////////////////////////////
//
//	BuildTestLevel()
//
void CwaWorld::BuildTestLevel(void)
{
	ResetLevel();

	DWORD i;

	for (DWORD south = 0; south < 40; ++south) {
		for (DWORD east = 0; east < 40; ++east) {
			m_Nodes[south][east].Flags   = CellFlag_Solid | CellFlag_HasFloor | CellFlag_HasCeiling;
			m_Nodes[south][east].Volume  = Texture_Wall;
			m_Nodes[south][east].Floor   = Texture_WoodPlank;
			m_Nodes[south][east].Ceiling = Texture_Marble;
			m_Nodes[south][east].Trigger = 0;
		}
	}

	m_Nodes[13][15].Flags &= ~CellFlag_Solid;
	m_Nodes[14][15].Flags &= ~CellFlag_Solid;
	m_Nodes[15][15].Flags &= ~CellFlag_Solid;
	m_Nodes[16][15].Flags &= ~CellFlag_Solid;

	m_Nodes[13][16].Flags &= ~CellFlag_Solid;
	m_Nodes[14][16].Flags &= ~CellFlag_Solid;
	m_Nodes[15][16].Flags &= ~CellFlag_Solid;
	m_Nodes[16][16].Flags &= ~CellFlag_Solid;

	m_Nodes[16][17].Flags &= ~CellFlag_Solid;
	m_Nodes[13][17].Flags &= ~CellFlag_Solid;

	m_Nodes[14][18].Flags &= ~CellFlag_Solid;
	m_Nodes[13][18].Flags &= ~CellFlag_Solid;
	m_Nodes[12][18].Flags &= ~CellFlag_Solid;

	m_Nodes[13][19].Flags &= ~CellFlag_Solid;
	m_Nodes[12][19].Flags &= ~CellFlag_Solid;

	m_Nodes[13][20].Flags &= ~CellFlag_Solid;

	m_Nodes[13][21].Flags &= ~CellFlag_Solid;

	m_Nodes[15][17].Flags &= ~CellFlag_HasFloor;
	m_Nodes[16][17].Flags &= ~CellFlag_HasFloor;
	m_Nodes[17][17].Flags &= ~CellFlag_HasFloor;

	m_Nodes[16][14].Volume = Texture_Mural1;
	m_Nodes[15][14].Volume = Texture_Mural2;
	m_Nodes[14][14].Volume = Texture_Mural3;
	m_Nodes[13][14].Volume = Texture_Cabinet;
	m_Nodes[12][15].Volume = Texture_GlassSword;
	m_Nodes[12][16].Volume = Texture_GlassDragon;
	m_Nodes[17][15].Volume = Texture_GlassHand;
	m_Nodes[17][16].Volume = Texture_GlassCrown;

	for (DWORD s =  7; s < 36; ++s) {
		for (DWORD e = 20; e < 30; ++e) {
			m_Nodes[s][e].Flags &= ~CellFlag_Solid;
		}
	}

	m_Nodes[10][18].Flags &= ~CellFlag_Solid;

	m_Nodes[11][18].Volume = Texture_GargoyleFace;
	m_Nodes[13][17].Volume = Texture_Door;

	SetCellTrigger(18, 11, CellFlag_SecretDoor);
	SetCellTrigger(17, 13, CellFlag_Door);


	m_MaxEast  = 30;
	m_MaxSouth = 36;


//	m_Position[0] = IntToFixed(22) + c_FixedPointHalf;
//	m_Position[1] = IntToFixed( 1) + c_EyeLevel;
//	m_Position[2] = IntToFixed(13) + c_FixedPointHalf;

	m_Position[0] = IntToFixed(15) + c_FixedPointHalf;
	m_Position[1] = IntToFixed( 1) + c_EyeLevel;
	m_Position[2] = IntToFixed(16) + c_FixedPointHalf;

	SetPlayerDirection(c_FixedPointHalf);


	const long separation = c_FixedPointOne + c_FixedPointHalf;

	DWORD lightCount = 0;

	for (i = 0; i < Animation_ArraySize; ++i) {
		DWORD id = g_AnimInfo[i].ID;

		m_Animations[id].LoadAnimation(g_AnimInfo[i].Name);

		m_Sprite[i].TextureID   = id;
		m_Sprite[i].Animated    = true;
		m_Sprite[i].Brightness  = 0;
		m_Sprite[i].BaseTime    = 0;
		m_Sprite[i].Width       = IntToFixed(m_Animations[id].m_Width)  / g_AnimInfo[i].Scale;
		m_Sprite[i].Height      = IntToFixed(m_Animations[id].m_Height) / g_AnimInfo[i].Scale;
		m_Sprite[i].Position[0] = IntToFixed(21) + ((i % 6) * separation);
		m_Sprite[i].Position[1] = IntToFixed( 1);
		m_Sprite[i].Position[2] = IntToFixed( 9) + ((i / 6) * separation);

		strcpy(m_Sprite[i].Name, g_AnimInfo[i].Name);

		if ((lightCount < ArraySize(m_Lights)) && (g_AnimInfo[i].Brightness > 0) && (i < 74)) {
			m_Lights[lightCount].Position[0]       = m_Sprite[i].Position[0];
			m_Lights[lightCount].Position[1]       = m_Sprite[i].Position[1];
			m_Lights[lightCount].Position[2]       = m_Sprite[i].Position[2];
			m_Lights[lightCount].BaseBrightness    = (g_AnimInfo[i].Brightness * c_FixedPointMask) / 100;

			++lightCount;
		}
	}


	for (i = 0; i < 25; ++i) {
		m_Creatures[i].LoadAnimation(g_CritterInfo[i].FileName);

		m_Critter[i].CritterID		= i;
		m_Critter[i].Direction		= 0;
		m_Critter[i].State			= 2;
		m_Critter[i].BaseTime		= 0;
		m_Critter[i].VocalizeTime	= 5000 + (rand() % 5000);
		m_Critter[i].Position[0]	= IntToFixed(21) + ((i % 6) * separation);
		m_Critter[i].Position[1]	= IntToFixed( 1);
		m_Critter[i].Position[2]	= IntToFixed(29) + ((i / 6) * separation);

		strcpy(m_Critter[i].Name, g_CritterInfo[i].DisplayName);
	}
}


/////////////////////////////////////////////////////////////////////////////
//
//	BuildTestCity()
//
void CwaWorld::BuildTestCity(void)
{
	memset(m_Sprite, 0, sizeof(m_Sprite));
	memset(m_Lights, 0, sizeof(m_Lights));
	memset(m_Projectile, 0, sizeof(m_Projectile));
}


/////////////////////////////////////////////////////////////////////////////
//
//	BuildDemoDungeon()
//
void CwaWorld::BuildDemoDungeon(void)
{
	memset(m_Sprite, 0, sizeof(m_Sprite));
	memset(m_Lights, 0, sizeof(m_Lights));
	memset(m_Projectile, 0, sizeof(m_Projectile));
}


